/**
 * Copyright (c) 2009-2012, Lukas Eder, lukas.eder@gmail.com
 * All rights reserved.
 *
 * This software is licensed to you under the Apache License, Version 2.0
 * (the "License"); You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * . Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * . Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * . Neither the name "jOOQ" nor the names of its contributors may be
 *   used to endorse or promote products derived from this software without
 *   specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.jooq;

import java.util.List;
import java.util.Map;

import org.jooq.exception.DataAccessException;

/**
 * Base functionality declaration for all query objects
 * <p>
 * This interface is for JOOQ INTERNAL USE only. Do not reference directly
 *
 * @author Lukas Eder
 */
public interface QueryPartInternal extends QueryPart {

    /**
     * Render this {@link QueryPart} to a SQL string contained in
     * <code>context.sql()</code>. The <code>context</code> will contain
     * additional information about how to render this <code>QueryPart</code>,
     * e.g. whether this <code>QueryPart</code> should be rendered as a
     * declaration or reference, whether this <code>QueryPart</code>'s contained
     * bind variables should be inlined or replaced by <code>'?'</code>, etc.
     */
    void toSQL(RenderContext context);

    /**
     * Retrieve the SQL that will be rendered by this {@link QueryPart}
     * <p>
     * This method is exposed publicly in {@link Query#getSQL()}
     */
    String getSQL();

    /**
     * Retrieve the SQL that will be rendered by this {@link QueryPart}
     * <p>
     * This method is exposed publicly in {@link Query#getSQL(boolean)}
     */
    String getSQL(boolean inline);

    /**
     * Retrieve the bind values that will be bound by this {@link QueryPart}
     * <p>
     * This method is exposed publicly in {@link Query#getBindValues()}
     */
    List<Object> getBindValues();

    /**
     * Retrieve the named parameters that will be bound by this {@link QueryPart}
     * <p>
     * This method is exposed publicly in {@link Query#getParams()}
     */
    Map<String, Param<?>> getParams();

    /**
     * Retrieve a named parameter that will be bound by this {@link QueryPart}
     * <p>
     * This method is exposed publicly in {@link Query#getParam(String)}
     */
    Param<?> getParam(String name);

    /**
     * Bind all parameters of this {@link QueryPart} to a PreparedStatement
     * <p>
     * This method is for JOOQ INTERNAL USE only. Do not reference directly
     *
     * @param context The context holding the next bind index and other
     *            information for variable binding
     * @throws DataAccessException If something went wrong while binding a
     *             variable
     */
    void bind(BindContext context) throws DataAccessException;

    /**
     * Reproduce the SQL dialect this {@link QueryPart} was created with
     * <p>
     * This method is for JOOQ INTERNAL USE only. Do not reference directly
     *
     * @return The SQL dialect
     * @deprecated - 2.0.2 - The attached SQL dialect of a {@link QueryPart}
     *             should no longer be referenced, as query parts become more
     *             and more {@link Configuration} - independent
     */
    @Deprecated
    SQLDialect getDialect();

    /**
     * Check whether this {@link QueryPart} is able to declare fields in a
     * <code>SELECT</code> clause.
     * <p>
     * This method can be used by any {@link Context} to check how a certain SQL
     * clause should be rendered.
     * <p>
     * This method is for JOOQ INTERNAL USE only. Do not reference directly
     */
    boolean declaresFields();

    /**
     * Check whether this {@link QueryPart} is able to declare tables in a
     * <code>FROM</code> clause or <code>JOIN</code> clause.
     * <p>
     * This method can be used by any {@link Context} to check how a certain SQL
     * clause should be rendered.
     * <p>
     * This method is for JOOQ INTERNAL USE only. Do not reference directly
     */
    boolean declaresTables();
}
